import gettext as gettextlib
import os
from collections.abc import Callable

import humanize.i18n

# Global translation function
_translator: Callable = gettextlib.gettext


LANG_MAP = {
    "ko_KR": "ko",
}


def init_translations(lang: str = "ko") -> None:
    """
    Initialize translations globally for the entire application.
    Call this once at application startup.
    """
    global _translator
    t_lang = LANG_MAP.get(lang, "en")
    localedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "locale")
    translations = gettextlib.translation("messages", localedir=localedir, languages=[t_lang], fallback=True)
    translations.install()
    _translator = translations.gettext


def gettext(message: str) -> str:
    """
    Get translated text. This function will use whatever language
    was set up by init_translations().
    """
    return _translator(message)


# Convenience alias that can be imported
_ = gettext


def set_locale(lang: str) -> None:
    """
    Set the locale for translations.
    """
    if lang not in ["en", "en_US", "en_GB"]:
        _ = humanize.i18n.activate(lang)
        _ = init_translations(lang=lang)



from ..models import TimeEntry
from .output import OutputFormatter


class RawTextOutput(OutputFormatter):
    def _entry_to_str(self, entry: TimeEntry) -> str:
        return f"{entry.id} {entry.start_time} {entry.end_time} {entry.project} {entry.comment} {entry.tags}"

    def single_entry_output(self, entry: TimeEntry | None) -> None:
        if entry is None:
            print("No active time record.")
            return
        print(self._entry_to_str(entry))

    def multiple_entries_output(self, entries: list[TimeEntry]) -> None:
        for entry in entries:
            print(self._entry_to_str(entry))



from datetime import datetime

import pytest
from freezegun import freeze_time

from ..datetime_utils import format_delta, parse_time


@freeze_time("2021-01-01")
def test_parse_time():
    # Test 24-hour format
    assert parse_time("12:00") == datetime(2021, 1, 1, 12, 0).astimezone()
    assert parse_time("23:59") == datetime(2021, 1, 1, 23, 59).astimezone()
    assert parse_time("00:00") == datetime(2021, 1, 1, 0, 0).astimezone()
    assert parse_time("01:02:03") == datetime(2021, 1, 1, 1, 2, 3).astimezone()

    # Test 12-hour format
    assert parse_time("12:00 PM") == datetime(2021, 1, 1, 12, 0).astimezone()
    assert parse_time("11:59 PM") == datetime(2021, 1, 1, 23, 59).astimezone()
    assert parse_time("12:00 AM") == datetime(2021, 1, 1, 0, 0).astimezone()
    assert parse_time("01:02:03 AM") == datetime(2021, 1, 1, 1, 2, 3).astimezone()
    assert parse_time("01:02:03 PM") == datetime(2021, 1, 1, 13, 2, 3).astimezone()

    # Test 12-hour format without leading zero
    assert parse_time("1:02:03 AM") == datetime(2021, 1, 1, 1, 2, 3).astimezone()
    assert parse_time("1:02:03 PM") == datetime(2021, 1, 1, 13, 2, 3).astimezone()

    # Test 12-hour format with meridiem in lowercase
    assert parse_time("12:00 pm") == datetime(2021, 1, 1, 12, 0).astimezone()
    assert parse_time("11:59 pm") == datetime(2021, 1, 1, 23, 59).astimezone()
    assert parse_time("12:00 am") == datetime(2021, 1, 1, 0, 0).astimezone()

    # Test invalid formats
    with pytest.raises(ValueError):
        parse_time("24:00")
    with pytest.raises(ValueError):
        parse_time("12:60")
    with pytest.raises(ValueError):
        parse_time("12:00:60")
    with pytest.raises(ValueError):
        parse_time("13:00 AM")
    with pytest.raises(ValueError):
        parse_time("13:00 PM")
    with pytest.raises(ValueError):
        parse_time("13:00:00 AM")
    with pytest.raises(ValueError):
        parse_time("13:00:00 PM")


def test_format_timedelta():
    assert format_delta(datetime(2021, 1, 1, 12, 0) - datetime(2021, 1, 1, 11, 0)) == "1 hour"
    assert format_delta(datetime(2021, 1, 1, 12, 0) - datetime(2021, 1, 1, 11, 30)) == "0.5 hours"
    assert format_delta(datetime(2021, 1, 1, 12, 0) - datetime(2021, 1, 1, 11, 45)) == "0.2 hours"



from enum import StrEnum

from ..models import TimeEntry


class OutputType(StrEnum):
    TEXT = "text"
    JSON = "json"
    RICH = "rich"
    YAML = "yaml"
    EMPTY = ""

    @classmethod
    def choices(cls) -> list[str]:
        return [x.value for x in cls]


class OutputFormatter:
    """The Base OutputFormatter class that all output classes should inherit from."""

    def __init__(self):
        pass

    def single_entry_output(self, entry: TimeEntry) -> None:
        raise NotImplementedError("output method is not implemented")

    def multiple_entries_output(self, entries: list[TimeEntry]) -> None:
        raise NotImplementedError("output method is not implemented")



from abc import ABC, abstractmethod

from ..models import EntryListFilter, TimeEntry


class TimeEntryRepository(ABC):
    @abstractmethod
    def save(self, entry: TimeEntry) -> TimeEntry:
        pass

    @abstractmethod
    def get_active_entry(self) -> TimeEntry | None:
        pass

    @abstractmethod
    def get_by_project(self, project: str) -> list[TimeEntry]:
        pass

    @abstractmethod
    def get_all(self) -> list[TimeEntry]:
        pass

    @abstractmethod
    def filter(self, *, filter: EntryListFilter):
        pass

    @abstractmethod
    def get_entry_by_id(self) -> TimeEntry:
        pass

    @abstractmethod
    def delete_entry(self, id: str) -> TimeEntry:
        pass



from datetime import date, datetime, timedelta
from typing import Literal
from uuid import uuid4

import humanize
import humanize.i18n
from pydantic import BaseModel, Field


class TimeEntry(BaseModel):
    id: str = Field(default_factory=lambda: uuid4().hex)
    start_time: datetime
    end_time: datetime | None = None
    project: str
    tags: set[str] = Field(default_factory=set)
    comment: str = ""

    @property
    def humanized_duration(self):
        td = self.end_time - self.start_time if self.end_time else datetime.now().astimezone() - self.start_time

        return humanize.precisedelta(
            td,
            suppress=("seconds", "milliseconds", "microseconds"),
            minimum_unit="hours",
            format="%0.1f",
        )

    @staticmethod
    def _get_naive_time(ts: datetime) -> datetime | None:
        """remove the tzinfo so we get the original hour"""
        return ts.replace(tzinfo=None) if ts else None

    @property
    def naive_start_time(self):
        return self._get_naive_time(self.start_time)

    @property
    def naive_end_time(self):
        return self._get_naive_time(self.end_time)

    def stop(self, end_time: datetime = None):
        if self.end_time is None or (self.end_time is not None and self.end_time < self.start_time):
            self.end_time = end_time or datetime.now().astimezone()
        else:
            raise ValueError("already stopped")

    @property
    def duration(self):
        return self.end_time - self.start_time if self.end_time else datetime.now().astimezone() - self.start_time


class EntryListFilter(BaseModel):
    id: str = ""
    projects: set[str] = Field(default_factory=set)
    start_date: date | None = None
    end_date: date | None = None
    tags: set[str] = Field(default_factory=set)
    time_period: Literal["today", "yesterday", "week", "month", "all", ""] | None = None

    def __init__(self, **data):
        super().__init__(**data)
        if self.time_period:
            self._apply_time_period()

    def _apply_time_period(self):
        """Apply date filters based on time period"""
        now = datetime.now()
        if self.time_period == "all":
            return  # No need to set date filters
        if self.time_period == "today":
            self.start_date = now.date()
        if self.time_period == "yesterday":
            self.start_date = now.date() - timedelta(days=1)
            self.end_date = now.date() - timedelta(days=1)
        elif self.time_period == "week":
            self.start_date = now.date() - timedelta(days=now.date().weekday())  # Monday
        elif self.time_period == "month":
            self.start_date = now.date() - timedelta(days=(now.date().day - 1))  # 1st of current month



# domain/services.py
import os
import subprocess
import tempfile
from datetime import datetime

import yaml

from .config.settings import Settings
from .models import EntryListFilter, TimeEntry
from .repositories.time_entry_repo import TimeEntryRepository
from .repositories.time_entry_repo_yaml import TimeEntryRepositoryYaml


class EditorServiceError(Exception): ...


class EditorService:
    """Base Editor Service class"""

    def __init__(self, editor_command: str):
        self.editor_command = editor_command

    def edit_entry(self, entry: TimeEntry) -> TimeEntry:
        raise NotImplementedError()


class YAMLEditorService(EditorService):
    """Editor service for YAML-formatted time entries"""

    def _format_entry_for_edit(self, entry: TimeEntry) -> str:
        """Format a time entry as YAML for editing"""
        entry_dict = entry.model_dump(mode="json")
        return yaml.dump(entry_dict)

    def _parse_edited_entry(self, content: str) -> TimeEntry:
        """Parse edited YAML content back into a TimeEntry"""
        try:
            data = yaml.safe_load(content)
            return TimeEntry(**data)
        except Exception as e:
            raise EditorServiceError(f"Invalid entry format: {str(e)}") from e

    def edit_entry(self, entry: TimeEntry) -> TimeEntry:
        """Edit a time entry in the user's preferred editor"""
        # Create temp file with entry content
        with tempfile.NamedTemporaryFile(suffix=".yaml", mode="w+", delete=False) as tmp:
            tmp.write(self._format_entry_for_edit(entry))
            tmp.flush()
            tmp_path = tmp.name

        try:
            # Open editor
            subprocess.run([self.editor_command, tmp_path], check=True)

            # Read and parse edited content
            with open(tmp_path) as f:
                edited_content = f.read()
                updated_entry = self._parse_edited_entry(edited_content)

            return updated_entry

        finally:
            # Clean up temp file
            os.unlink(tmp_path)


class TimeTrackingService:
    def __init__(
        self,
        settings: Settings,
        repository: TimeEntryRepository | None = None,
        editor: EditorService | None = None,
    ):
        self.settings = settings
        self.repository = repository or TimeEntryRepositoryYaml(settings.data_filename)
        self.editor = editor or YAMLEditorService(settings.editor)

    def start_tracking(
        self,
        project: str,
        start_time: datetime | None = None,
        comment: str = "",
        tags: set[str] = None,
    ) -> TimeEntry:
        start_time = start_time or datetime.now().astimezone()
        # Stop any active entry first
        if active_entry := self.repository.get_active_entry():
            active_entry.stop(end_time=start_time)
            self.repository.save(active_entry)

        # Apply auto-tagging rules
        auto_tags = self.settings.apply_auto_tags(project)
        final_tags = (tags or set()) | auto_tags

        # Create and save new entry
        new_entry = TimeEntry(
            project=project,
            start_time=start_time,
            comment=comment,
            tags=final_tags,
        )
        self.repository.save(new_entry)
        return new_entry

    def stop_tracking(self, stop_time: datetime | None = None) -> TimeEntry | None:
        active_entry = self.repository.get_active_entry()
        if active_entry:
            active_entry.stop(stop_time)
            self.repository.save(active_entry)
            return active_entry
        return None

    def get_active_entry(self) -> TimeEntry | None:
        return self.repository.get_active_entry()

    def list_entries(self, filter: EntryListFilter | None = None) -> list[TimeEntry]:
        if filter:
            return self.repository.filter(filter=filter)
        return self.repository.get_all()

    def get_entry(self, id: str) -> TimeEntry:
        return self.repository.get_entry_by_id(id)

    def update_entry(self, entry: TimeEntry) -> TimeEntry:
        """Update an existing time entry"""
        # Verify the entry exists first
        self.get_entry(entry.id)
        self.repository.save(entry)
        return entry

    def delete_entry(self, id: str):
        return self.repository.delete_entry(id)

    def edit_entry(self, id: str):
        entry = self.get_entry(id)
        updated_entry = self.editor.edit_entry(entry)
        return self.update_entry(updated_entry)

    def get_entry_by_partial_id(self, partial_id: str) -> TimeEntry:
        entries = self.list_entries(EntryListFilter(id=partial_id))
        if len(entries) == 1:
            return entries[0]
        elif len(entries) > 1:
            raise IndexError("Multiple entries found")
        raise KeyError("record id not found")



from .. import create_output_formatter
from ..json_output import JsonOutput
from ..output import OutputFormatter
from ..rich_text_output import RichTextOutput
from ..text_output import RawTextOutput


def test_output():
    """Test the output module"""
    output = OutputFormatter()
    assert output

    output = create_output_formatter("rich", force=True)
    assert isinstance(output, RichTextOutput)

    output = create_output_formatter("text", force=True)
    assert isinstance(output, RawTextOutput)

    output = create_output_formatter("json", force=True)
    assert isinstance(output, JsonOutput)



import sys

from .json_output import JsonOutput
from .output import OutputFormatter, OutputType
from .output_utils import validate_output_format
from .rich_text_output import RichTextOutput
from .text_output import RawTextOutput
from .yaml_output import YamlOutput

__all__ = [
    JsonOutput,
    OutputFormatter,
    OutputType,
    RawTextOutput,
    RichTextOutput,
    YamlOutput,
    validate_output_format,
    "create_output_formatter",
]


def create_output_formatter(output_format: OutputType | None, force: bool = False) -> OutputFormatter:
    """Function to create an output formatter based on the output format."""
    if not force:
        if sys.stdout.isatty() and output_format is None or output_format in (OutputType.EMPTY, OutputType.RICH):
            output_format = OutputType.RICH
        elif not sys.stdout.isatty() and output_format is None or output_format in (OutputType.EMPTY, OutputType.JSON):
            output_format = OutputType.JSON
        else:
            output_format = OutputType.TEXT
    match output_format:
        case OutputType.TEXT:
            return RawTextOutput()
        case OutputType.JSON:
            return JsonOutput()
        case OutputType.RICH:
            return RichTextOutput()
        case OutputType.YAML:
            return YamlOutput()
    raise ValueError(f"Unsupported output format: {output_format}")



from datetime import date, timedelta

from rich.console import Console
from rich.table import Table

from ..models import TimeEntry
from ..utils.datetime_utils import format_delta
from ..utils.translation import gettext as _
from .output import OutputFormatter

ABBR_ID_LENGTH = 4


class RichTextOutput(OutputFormatter):
    def single_entry_output(self, entry: TimeEntry | None) -> None:
        console = Console()
        if entry is None:
            console.print("🕛 " + _("No active time record."), style="yellow")
            return
        table = Table(title=_("Time Record"))
        table.add_column(_("field"))
        table.add_column(_("value"))
        table.add_row(
            _("ID"),
            f"[magenta]{entry.id[0:ABBR_ID_LENGTH]}[/magenta]{entry.id[ABBR_ID_LENGTH:]}",
        )
        table.add_row(_("start time"), f"[cyan]{entry.naive_start_time:%Y-%m-%d %H:%M:%S}[/cyan]")
        table.add_row(
            _("end time"),
            f"[magenta]{entry.naive_end_time:%H:%M:%S}[/magenta]" if entry.end_time else "-",
        )
        table.add_row(_("delta"), f"[cyan]{entry.humanized_duration}[/cyan]")
        table.add_row(_("project"), f"[green]{entry.project}[/green]")
        table.add_row(_("comments"), f"[blue]{entry.comment}[/blue]")
        table.add_row(_("tags"), "[red]" + ", ".join(tag for tag in entry.tags) + "[/red]")
        console = Console()
        console.print(table)

    def multiple_entries_output(self, entries: list[TimeEntry]) -> None:
        def _check_and_print_total(table: Table, td: timedelta, description: str, style: str):
            if td:
                table.add_row(
                    "",
                    description,
                    "",
                    format_delta(td),
                    style=style,
                )

        table = Table(title=_("Time Records"))
        table.add_column(_("id"), justify="left", style="#707070")
        table.add_column(_("start"), justify="right", style="cyan")
        table.add_column(_("end"), justify="right", style="magenta")
        table.add_column(_("delta"), style="cyan")
        table.add_column(_("project"), justify="left", style="green")
        table.add_column(_("comments"), style="blue")
        table.add_column(_("tags"), style="red")
        current_date = date(1970, 1, 1)
        subtotal_delta = timedelta(0)
        total_delta = timedelta(0)
        for entry in entries:
            if entry.start_time.date() != current_date:
                _check_and_print_total(table, subtotal_delta, _("subtotal"), "#a0a0a0")
                subtotal_delta = timedelta(0)
                current_date = entry.start_time.date()
                table.add_section()
                table.add_row("", f"{current_date:%Y-%m-%d}", style="yellow")
            table.add_row(
                f"{entry.id[0:ABBR_ID_LENGTH]}",
                f"{entry.naive_start_time:%H:%M:%S}",
                f"{entry.naive_end_time:%H:%M:%S}" if entry.end_time else "-",
                f"{entry.humanized_duration}",
                entry.project,
                entry.comment,
                ", ".join(tag for tag in entry.tags),
            )
            subtotal_delta += entry.duration
            total_delta += entry.duration

        _check_and_print_total(table, subtotal_delta, _("subtotal"), "#a0a0a0")
        table.add_section()
        _check_and_print_total(table, total_delta, _("total"), "yellow")
        console = Console()
        console.print(table)



import re
from datetime import datetime, time, timedelta

import click
import humanize


def parse_time(time_str: str) -> datetime:
    """Parse user-entered time string into datetime object for current date."""
    time_str = time_str.strip().upper()

    pattern = r"^(\d{1,2}):(\d{2})(?::(\d{2}))?\s*((?:am|pm)?)"
    match = re.match(pattern, time_str, re.IGNORECASE)

    if not match:
        raise ValueError("Invalid time format. Use HH:MM or HH:MM:SS with optional AM/PM")

    hours, minutes, seconds, meridiem = match.groups()
    hours = int(hours)
    minutes = int(minutes)
    seconds = int(seconds) if seconds else 0

    if hours > 12 and meridiem:
        raise ValueError("Hours cannot exceed 12 in 12-hour format")

    meridiem = meridiem.upper() if meridiem else ""

    if meridiem == "PM" and hours < 12:
        hours += 12
    elif meridiem == "AM" and hours == 12:
        hours = 0

    if hours > 23:
        raise ValueError("Hours cannot exceed 23")
    if minutes > 59:
        raise ValueError("Minutes cannot exceed 59")
    if seconds > 59:
        raise ValueError("Seconds cannot exceed 59")

    today = datetime.now().date()
    return datetime.combine(today, time(hours, minutes, seconds)).astimezone()


def validate_time(ctx: click.Context, param: click.Parameter, value: str | None) -> datetime | None:
    if value is None:
        return None
    try:
        return parse_time(value)
    except ValueError as ex:
        raise click.BadParameter(str(ex)) from ex


def format_delta(td: timedelta) -> str:
    return humanize.precisedelta(
        td,
        suppress=(
            "years",
            "months",
            "days",
        ),
        minimum_unit="hours",
        format="%0.1f",
    )



import os

import yaml

from ..models import EntryListFilter, TimeEntry
from .time_entry_repo import TimeEntryRepository


class TimeEntryRepositoryYaml(TimeEntryRepository):
    def __init__(self, filename: str = "timesheet.yaml"):
        self.filename = filename
        self._cache = None
        self._ensure_file_exists()

    def _ensure_file_exists(self):
        if not os.path.exists(self.filename):
            self._save_data({"entries": []})

    def _load_data(self) -> dict:
        if self._cache is None:
            with open(self.filename) as f:
                self._cache = yaml.load(f, yaml.FullLoader)
        return self._cache

    def _save_data(self, data: dict):
        with open(self.filename, "w") as f:
            yaml.dump(data, f)
        self._cache = data

    def _invalidate_cache(self):
        self._cache = None

    def get_active_entry(self) -> TimeEntry | None:
        data = self._load_data()
        active = next((e for e in data["entries"] if not e.get("end_time")), None)
        return TimeEntry(**active) if active else None

    def get_all(self) -> list[TimeEntry]:
        data = self._load_data()
        # Since entries are maintained in sorted order during save operations,
        # we don't need to sort here anymore
        return [TimeEntry(**entry) for entry in data["entries"]]

    def get_by_project(self, project: str) -> list[TimeEntry]:
        data = self._load_data()
        return [TimeEntry(**entry) for entry in data["entries"] if entry["project"] == project]

    def get_entry_by_id(self, id: str) -> TimeEntry:
        data = self._load_data()
        for entry in data["entries"]:
            if entry["id"] == id:
                return TimeEntry(**entry)
        raise KeyError("record id not found")

    def delete_entry(self, id: str) -> TimeEntry:
        data = self._load_data()
        entries = []
        found = None
        for entry in data["entries"]:
            if entry["id"] == id:
                found = entry
            else:
                entries.append(entry)
        if not found:
            raise KeyError("record id not found")
        self._save_data({"entries": entries})
        return TimeEntry(**found)

    @staticmethod
    def _project_matching(filter_projects: set[str], project: str) -> bool:
        for f_proj in filter_projects:
            if (f_proj[-1] in "*+." and project.startswith(f_proj[0:-1])) or f_proj == project:
                return True
        return False

    def _check_against_filter(self, filter: EntryListFilter, entry: TimeEntry) -> bool:
        conditions = [
            # id filter
            (not filter.id or entry.id.startswith(filter.id)),
            # Project filters
            (not filter.projects or self._project_matching(filter.projects, entry.project)),
            # Date filters
            (not filter.start_date or entry.start_time.date() >= filter.start_date),
            (not filter.end_date or (entry.end_time and entry.end_time.date() <= filter.end_date)),
            # Tag filter
            (not filter.tags or any(tag in entry.tags for tag in filter.tags)),
        ]
        return all(conditions)

    def filter(self, *, filter: EntryListFilter | None = None) -> list[TimeEntry]:
        time_entries = self.get_all()
        if filter is None:
            return time_entries
        return [entry for entry in time_entries if self._check_against_filter(filter, entry)]

    def save(self, entry: TimeEntry) -> None:
        data = self._load_data()
        entry_dict = entry.model_dump(mode="json")
        found = False
        entries = []
        for e in data["entries"]:
            if e["id"] == entry.id:
                entries.append(entry_dict)
                found = True
            else:
                entries.append(e)
        if not found:
            entries.append(entry_dict)

        # Sort entries by start_time before saving
        entries.sort(key=lambda x: x["start_time"])
        data["entries"] = entries
        self._save_data(data)



from datetime import datetime, timedelta

import pytest

from ..models import EntryListFilter, TimeEntry


def test_time_entry():
    now = datetime.now().astimezone()

    t1 = TimeEntry(start_time=now + timedelta(hours=-4), project="test-project", comment="hello")
    t2 = TimeEntry(
        start_time=now + timedelta(hours=-2, minutes=-30),
        end_time=now,
        project="test-project-2",
        comment="test2",
        tags=["testtag2"],
    )

    assert t1.humanized_duration == "4 hours"
    assert t2.humanized_duration == "2.5 hours"  # our current precisiondelta config


def test_time_entry_stop():
    now = datetime.now().astimezone()

    t1 = TimeEntry(start_time=now + timedelta(hours=-4), project="test-project", comment="hello")
    t1.stop()
    assert t1.humanized_duration == "4 hours"
    assert t1.end_time is not None
    # verify we can't re-stop an already stopped entry
    with pytest.raises(ValueError):
        t1.stop()


def test_get_naive_time():
    now = datetime.now().astimezone()
    naive_time = TimeEntry._get_naive_time(now)
    assert naive_time.tzinfo is None
    assert naive_time.hour == now.hour


def test_duration():
    now = datetime.now().astimezone()
    t2 = TimeEntry(
        start_time=now + timedelta(hours=-2),
        end_time=now,
        project="test-project-2",
        comment="test2",
        tags=["testtag2"],
    )
    assert t2.duration == t2.end_time - t2.start_time


def test_entry_list_filter_time_period_today():
    now = datetime.now()
    filter = EntryListFilter(time_period="today")
    assert filter.start_date == now.date()
    assert filter.end_date is None


def test_entry_list_filter_time_period_week():
    now = datetime.now()
    filter = EntryListFilter(time_period="week")
    assert filter.start_date == now.date() - timedelta(days=now.date().weekday())
    assert filter.end_date is None


def test_entry_list_filter_time_period_month():
    now = datetime.now()
    filter = EntryListFilter(time_period="month")
    assert filter.start_date == now.date() - timedelta(days=(now.date().day - 1))
    assert filter.end_date is None


def test_entry_list_filter_projects():
    filter = EntryListFilter(projects={"project1", "project2"})
    assert len(filter.projects) == 2
    assert "project1" in filter.projects
    assert "project2" in filter.projects


def test_entry_list_filter_tags():
    filter = EntryListFilter(tags={"tag1", "tag2"})
    assert len(filter.tags) == 2
    assert "tag1" in filter.tags
    assert "tag2" in filter.tags



from unittest import mock

from click.testing import CliRunner

from ..cli import cli, load_settings
from ..models import TimeEntry
from ..output.text_output import RawTextOutput


def mock_single_entry_output(entry: TimeEntry):
    print(f"{entry.id} {entry.start_time} {entry.end_time} {entry.project} {entry.comment}")


def test_load_settings(tmp_path):
    """Test settings loading with default and custom config"""
    settings = load_settings()
    assert settings is not None

    # Test with custom config
    config_file = tmp_path / "config.yaml"
    config_content = """
    locale: en_US
    data_filename: test.yaml
    """
    config_file.write_text(config_content)
    settings = load_settings(config_file)
    assert settings.locale == "en_US"
    assert settings.data_filename == "test.yaml"


def test_start_command(tmp_path):
    """Test the start command"""
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # Basic start command
        result = runner.invoke(cli, ["-f", "test.yaml", "start", "test-project"])
        assert result.exit_code == 0
        assert "test-project" in result.output

        # Start with comment and tags
        result = runner.invoke(
            cli,
            [
                "-f",
                "test.yaml",
                "start",
                "test-project",
                "test comment",
                "--tag",
                "tag1",
                "--tag",
                "tag2",
            ],
        )
        assert result.exit_code == 0
        assert "test-project" in result.output
        assert "test comment" in result.output
        assert "tag1" in result.output
        assert "tag2" in result.output

        # Start with specific time
        result = runner.invoke(cli, ["-f", "test.yaml", "start", "test-project", "--start_time", "09:00"])
        assert result.exit_code == 0
        assert "09:00" in result.output

        # Invalid time format
        result = runner.invoke(cli, ["-f", "test.yaml", "start", "test-project", "--start_time", "invalid"])
        assert result.exit_code != 0
        assert "Invalid time format" in result.output


def test_stop_command(tmp_path):
    """Test the stop command"""
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # Start tracking first
        runner.invoke(cli, ["-f", "test.yaml", "start", "test-project"])

        # Basic stop
        result = runner.invoke(cli, ["-f", "test.yaml", "stop"])
        assert result.exit_code == 0
        assert "test-project" in result.output

        # Stop with specific time
        runner.invoke(cli, ["-f", "test.yaml", "start", "test-project"])
        result = runner.invoke(cli, ["-f", "test.yaml", "stop", "--stop_time", "17:00"])
        assert result.exit_code == 0
        assert "17:00" in result.output

        # Invalid time format
        result = runner.invoke(cli, ["-f", "test.yaml", "stop", "--stop_time", "invalid"])
        assert result.exit_code != 0
        assert "Invalid time format" in result.output


def test_status_command(tmp_path):
    """Test the status command"""
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # Check status with no active tracking
        result = runner.invoke(cli, ["-f", "test.yaml", "status"])
        assert result.exit_code == 0
        assert "No active time record" in result.output

        # Start tracking and check status
        runner.invoke(cli, ["-f", "test.yaml", "start", "test-project", "test comment"])
        result = runner.invoke(cli, ["-f", "test.yaml", "status"])
        assert result.exit_code == 0
        assert "test-project" in result.output
        assert "test comment" in result.output


def test_list_command_and_aliases(tmp_path):
    """Test the list command and its aliases"""
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # Create some entries
        runner.invoke(cli, ["-f", "test.yaml", "start", "project1", "--tag", "tag1"])
        runner.invoke(cli, ["-f", "test.yaml", "stop"])
        runner.invoke(cli, ["-f", "test.yaml", "start", "project2", "--tag", "tag2"])
        runner.invoke(cli, ["-f", "test.yaml", "stop"])

        # Test main 'list' command
        result = runner.invoke(cli, ["-f", "test.yaml", "list"])
        assert result.exit_code == 0
        assert "project1" in result.output
        assert "project2" in result.output

        # Test 'ls' alias
        result = runner.invoke(cli, ["-f", "test.yaml", "ls"])
        assert result.exit_code == 0
        assert "project1" in result.output
        assert "project2" in result.output

        # List with time period
        result = runner.invoke(cli, ["-f", "test.yaml", "list", "today"])
        assert result.exit_code == 0
        assert "project1" in result.output
        assert "project2" in result.output

        # List with project filter
        result = runner.invoke(cli, ["-f", "test.yaml", "list", "--project", "project1"])
        assert result.exit_code == 0
        assert "project1" in result.output
        assert "project2" not in result.output

        # List with tag filter
        result = runner.invoke(cli, ["-f", "test.yaml", "list", "--tag", "tag1"])
        assert result.exit_code == 0
        assert "project1" in result.output
        assert "project2" not in result.output


def test_edit_command(tmp_path):
    """Test the edit command"""
    runner = CliRunner()
    with (
        mock.patch("sigye.output.rich_text_output.RichTextOutput", side_effect=RawTextOutput),
        runner.isolated_filesystem(temp_dir=tmp_path),
    ):
        # Create an entry
        result = runner.invoke(cli, ["-f", "test.yaml", "start", "test-project"])
        runner.invoke(cli, ["-f", "test.yaml", "stop"])

        # Extract the ID from the output
        entry_id = result.output.split()[0]  # Assuming ID is first word in output
        assert entry_id is not None

        # Try to edit with invalid ID
        result = runner.invoke(cli, ["-f", "test.yaml", "edit", "invalid-id"])
        assert result.exit_code != 0
        assert "No entry found" in result.output


def test_delete_command_and_aliases(tmp_path):
    """Test the delete command and its aliases"""
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # Create an entry
        result = runner.invoke(cli, ["-f", "test.yaml", "-o", "text", "start", "test-project"])
        runner.invoke(cli, ["-f", "test.yaml", "-o", "text", "stop"])
        print(f"output: {result.output}")
        # Extract the ID from the output
        entry_id = result.output.split()[0]  # Assuming ID is first word in output

        # Test main 'delete' command
        result = runner.invoke(cli, ["-f", "test.yaml", "-o", "text", "delete", entry_id])
        assert result.exit_code == 0
        assert "test-project" in result.output

        # Create another entry for testing aliases
        result = runner.invoke(cli, ["-f", "test.yaml", "-o", "text", "start", "test-project"])
        runner.invoke(cli, ["-f", "test.yaml", "stop"])
        entry_id = result.output.split()[0]

        # Test 'rm' alias
        result = runner.invoke(cli, ["-f", "test.yaml", "-o", "text", "rm", entry_id])
        assert result.exit_code == 0
        assert "test-project" in result.output

        # Create another entry for testing aliases
        result = runner.invoke(cli, ["-f", "test.yaml", "-o", "text", "start", "test-project"])
        runner.invoke(cli, ["-f", "test.yaml", "stop"])
        entry_id = result.output.split()[0]

        # Test 'del' alias
        result = runner.invoke(cli, ["-f", "test.yaml", "-o", "text", "del", entry_id])
        assert result.exit_code == 0
        assert "test-project" in result.output

        # Try to delete with invalid ID
        result = runner.invoke(cli, ["-f", "test.yaml", "-o", "text", "delete", "invalid-id"])
        assert result.exit_code != 0
        assert "No entry found" in result.output


def test_config_file_option(tmp_path):
    """Test using custom config file"""
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # Create custom config
        config_file = tmp_path / "custom_config.yaml"
        config_content = """
        locale: en_US
        data_filename: custom_test.yaml
        """
        config_file.write_text(config_content)

        # Use custom config
        result = runner.invoke(cli, ["-c", str(config_file), "start", "test-project"])
        assert result.exit_code == 0
        assert "test-project" in result.output



from datetime import datetime, timedelta
from pathlib import Path

import pytest

from ..config.settings import AutoTagRule, Settings
from ..models import EntryListFilter
from ..repositories.time_entry_repo_yaml import TimeEntryRepositoryYaml
from ..services import EditorService, TimeTrackingService, YAMLEditorService


class DummyEditorService(EditorService):
    def edit_entry(self, entry):
        entry.comment = "edited by dummy editor"
        return entry


def create_test_settings(tmp_path: Path) -> Settings:
    """Create test settings with a specific configuration"""
    settings = Settings(data_filename=str(tmp_path / "test.yaml"), locale="en_US")
    settings.auto_tag_rules = [
        AutoTagRule(pattern="^abc", match_type="regex", tags=["learning"]),
        AutoTagRule(pattern="^PROJ-\\d+", match_type="regex", tags=["work", "billable"]),
        AutoTagRule(pattern="(feature|bugfix)/", match_type="regex", tags=["development"]),
        AutoTagRule(pattern="-urgent$", match_type="regex", tags=["priority"]),
    ]
    return settings


def test_initialization(tmp_path):
    settings = create_test_settings(tmp_path)
    tts = TimeTrackingService(settings=settings)
    assert tts.settings == settings
    assert tts.repository.filename == tts.settings.data_filename
    assert isinstance(tts.editor, YAMLEditorService)
    tts = TimeTrackingService(settings=settings, editor=DummyEditorService("nothing"))
    assert isinstance(tts.editor, DummyEditorService)


def test_basic_time_tracking(tmp_path):
    filename = tmp_path / "test.yaml"
    settings = create_test_settings(tmp_path)
    tts = TimeTrackingService(repository=TimeEntryRepositoryYaml(filename), settings=settings)
    assert tts.list_entries() == []
    a = tts.start_tracking("test-project")
    assert a.end_time is None
    assert tts.get_active_entry() == a
    b = tts.stop_tracking()
    assert b.start_time == a.start_time
    assert b.end_time is not None
    c = tts.stop_tracking()
    assert c is None

    entries = tts.list_entries()
    assert len(entries) == 1
    assert entries[0] == b


def test_auto_tagging(tmp_path):
    filename = tmp_path / "test.yaml"
    settings = create_test_settings(tmp_path)
    tts = TimeTrackingService(repository=TimeEntryRepositoryYaml(filename), settings=settings)

    # Test regex matching with '^abc' pattern
    entry1 = tts.start_tracking("abc-learning-task")
    assert "learning" in entry1.tags
    tts.stop_tracking()

    # Test regex matching with '^PROJ-\d+' pattern
    entry2 = tts.start_tracking("PROJ-123")
    assert "work" in entry2.tags
    assert "billable" in entry2.tags
    tts.stop_tracking()

    # Test regex matching with '(feature|bugfix)/' pattern
    entry3 = tts.start_tracking("feature/new-ui")
    assert "development" in entry3.tags
    tts.stop_tracking()

    entry4 = tts.start_tracking("bugfix/issue-123")
    assert "development" in entry4.tags
    tts.stop_tracking()

    # Test regex matching with '-urgent$' pattern
    entry5 = tts.start_tracking("task-urgent")
    assert "priority" in entry5.tags
    tts.stop_tracking()

    # Test no auto-tagging for non-matching project
    entry6 = tts.start_tracking("regular-task")
    assert len(entry6.tags) == 0
    tts.stop_tracking()

    # Test merging auto-tags with user-provided tags
    entry7 = tts.start_tracking("abc-task", tags={"custom-tag"})
    assert "learning" in entry7.tags
    assert "custom-tag" in entry7.tags
    tts.stop_tracking()


def test_start_stop(tmp_path):
    filename = tmp_path / "test.yaml"
    settings = create_test_settings(tmp_path)
    tts = TimeTrackingService(repository=TimeEntryRepositoryYaml(filename), settings=settings)

    d1 = tts.start_tracking("test-project")
    assert d1.end_time is None
    d2 = tts.start_tracking("test-project-2")
    assert d2.end_time is None
    d1 = tts.get_entry(d1.id)
    assert d1.end_time is not None


def test_filtered_list(tmp_path):
    filename = tmp_path / "test.yaml"
    settings = create_test_settings(tmp_path)
    tts = TimeTrackingService(repository=TimeEntryRepositoryYaml(filename), settings=settings)

    # Create entries with different projects and tags
    d1 = tts.start_tracking("test-project", tags={"tag1", "tag2"})
    tts.stop_tracking()
    d2 = tts.start_tracking("test-project-2", tags={"tag2", "tag3"})
    tts.stop_tracking()
    d3 = tts.start_tracking("test-project-3", tags={"tag1", "tag3"})
    tts.stop_tracking()
    assert all(entry is not None for entry in [d1, d2, d3])

    # Test no filter
    l1 = tts.list_entries(filter=None)
    assert len(l1) == 3

    # Test project filter
    filter_by_projects = EntryListFilter(projects={"test-project", "test-project-2"})
    l2 = tts.list_entries(filter=filter_by_projects)
    assert len(l2) == 2
    assert all(e.project in {"test-project", "test-project-2"} for e in l2)

    # Test tag filter
    filter_by_tags = EntryListFilter(tags={"tag1"})
    l3 = tts.list_entries(filter=filter_by_tags)
    assert len(l3) == 2
    assert all("tag1" in e.tags for e in l3)

    # Test date filter
    now = datetime.now()
    yesterday = now - timedelta(days=1)
    filter_by_date = EntryListFilter(start_date=yesterday.date())
    l5 = tts.list_entries(filter=filter_by_date)
    assert len(l5) == 3

    # Test combined filters
    combined_filter = EntryListFilter(projects={"test-project"}, tags={"tag1"}, start_date=yesterday.date())
    l6 = tts.list_entries(filter=combined_filter)
    assert len(l6) == 1
    assert l6[0].project == "test-project"
    assert "tag1" in l6[0].tags

    # Test time period filter
    period_filter = EntryListFilter(time_period="today")
    l7 = tts.list_entries(filter=period_filter)
    assert len(l7) == 3
    assert all(e.start_time.date() == now.date() for e in l7)


def test_invalid_entry_fetch(tmp_path):
    filename = tmp_path / "test.yaml"
    settings = create_test_settings(tmp_path)
    tts = TimeTrackingService(repository=TimeEntryRepositoryYaml(filename), settings=settings)

    d1 = tts.start_tracking("test-project")
    assert d1.end_time is None
    tts.stop_tracking()
    d1 = tts.get_entry(d1.id)
    assert d1.end_time is not None

    with pytest.raises(KeyError):
        _ = tts.get_entry("invalid value here!")


def test_delete_entry(tmp_path):
    filename = tmp_path / "test.yaml"
    settings = create_test_settings(tmp_path)
    tts = TimeTrackingService(repository=TimeEntryRepositoryYaml(filename), settings=settings)

    d1 = tts.start_tracking("test-project")
    assert d1.end_time is None
    tts.stop_tracking()
    d1 = tts.get_entry(d1.id)
    assert d1.end_time is not None

    tts.delete_entry(d1.id)

    with pytest.raises(KeyError):
        _ = tts.get_entry(d1.id)


def test_configuration_loading(tmp_path):
    """Test that configuration is properly loaded and applied"""
    # Create a config file in the test directory
    config_file = tmp_path / ".sigye" / "config.yaml"
    config_file.parent.mkdir(parents=True, exist_ok=True)
    config_content = f"""
locale: ko_KR
data_filename: {str(tmp_path / "test.yaml")}
auto_tag_rules:
  - pattern: "^test-"
    match_type: "regex"
    tags: ["testing"]
"""
    config_file.write_text(config_content)

    # Create settings with the test directory as home
    settings = Settings.load_from_file(config_file)

    assert settings.locale == "ko_KR"  # Verify locale override

    # Create service and test auto-tagging
    tts = TimeTrackingService(repository=TimeEntryRepositoryYaml(settings.data_filename), settings=settings)

    entry = tts.start_tracking("test-project")
    assert "testing" in entry.tags


def test_edit_command(tmp_path):
    settings = create_test_settings(tmp_path)
    tts = TimeTrackingService(settings=settings, editor=DummyEditorService("nothing"))

    d1 = tts.start_tracking("test-project")
    assert d1.end_time is None
    d1 = tts.edit_entry(d1.id)
    d1 = tts.get_entry(d1.id)
    assert d1.comment == "edited by dummy editor"



from datetime import datetime, timedelta

from ...models import TimeEntry
from ..rich_text_output import RichTextOutput


def test_rich_text_output():
    now = datetime.now().astimezone()

    t1 = TimeEntry(start_time=now + timedelta(hours=-4), project="test-project", comment="hello")
    t2 = TimeEntry(
        start_time=now + timedelta(hours=-2, minutes=-30),
        end_time=now,
        project="test-project-2",
        comment="test2",
        tags=["testtag2"],
    )

    output = RichTextOutput()
    output.single_entry_output(t1)
    output.single_entry_output(t2)
    output.multiple_entries_output([t1, t2])

    # TODO: add assertions for the output



import json

from ..models import TimeEntry
from .output import OutputFormatter


class JsonOutput(OutputFormatter):
    def single_entry_output(self, entry: TimeEntry | None) -> None:
        if entry is None:
            return
        print(json.dumps(entry.model_dump(mode="json")))

    def multiple_entries_output(self, entries: list[TimeEntry]) -> None:
        print(json.dumps([entry.model_dump(mode="json") for entry in entries]))



import os
import re
from pathlib import Path
from typing import Literal, Self

import yaml
from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from ..output import OutputType

DEFAULT_HOME_DIRECTORY = Path.home() / ".sigye"
DEFAULT_HOME_DIRECTORY.mkdir(exist_ok=True, parents=True)
DEFAULT_CONFIG_PATH = DEFAULT_HOME_DIRECTORY / "config.yaml"
DEFAULT_DATA_FILENAME = DEFAULT_HOME_DIRECTORY / "time_entries.yml"
DEFAULT_EDITOR = os.getenv("EDITOR", "vim")


class AutoTagRule(BaseModel):
    pattern: str
    match_type: Literal["regex"]  # Using regex for all pattern matching
    tags: list[str]


class Settings(BaseSettings):
    data_filename: str = Field(default=str(DEFAULT_DATA_FILENAME))
    locale: str = Field(default="en_US")
    auto_tag_rules: list[AutoTagRule] = []
    editor: str = Field(default=DEFAULT_EDITOR)
    output_format: OutputType = Field(default=OutputType.EMPTY)

    @classmethod
    def load_from_file(cls, path: Path = DEFAULT_CONFIG_PATH) -> Self:
        if not path.exists():
            return cls()
        with open(path) as f:
            config_dict = yaml.safe_load(f)
            return cls.model_validate(config_dict or {})

    def apply_auto_tags(self, project: str) -> set[str]:
        """Apply auto-tagging rules to a project name and return the set of tags to add"""
        tags = set()
        for rule in self.auto_tag_rules:
            if re.search(rule.pattern, project):
                tags.update(rule.tags)
        return tags

    model_config = SettingsConfigDict(env_prefix="sigye_", extra="ignore")



from dataclasses import dataclass
from pathlib import Path

import click
from click_aliases import ClickAliasedGroup
from click_datetime import Datetime

from .config.settings import DEFAULT_CONFIG_PATH, Settings
from .models import EntryListFilter
from .output import OutputFormatter, OutputType, create_output_formatter, validate_output_format
from .services import EditorServiceError, TimeTrackingService
from .utils.datetime_utils import validate_time
from .utils.translation import set_locale


def load_settings(config_file: Path | None = None) -> Settings:
    """Load settings from config file, falling back to defaults if needed"""
    settings = Settings.load_from_file(config_file) if config_file else Settings.load_from_file()
    set_locale(settings.locale)
    return settings


@dataclass
class ContextObject:
    tts: TimeTrackingService
    output: OutputFormatter


pass_context_object = click.make_pass_decorator(ContextObject, ensure=True)


@click.group(cls=ClickAliasedGroup)
@click.version_option()
@click.option(
    "--config-file",
    "-c",
    type=click.Path(path_type=Path),
    default=DEFAULT_CONFIG_PATH,
    help="Path to config file",
    show_default=True,
)
@click.option(
    "--filename",
    "-f",
    type=click.Path(path_type=Path),
    help="Path to data file",
)
@click.option(
    "--output_format",
    "-o",
    type=click.Choice(OutputType.choices(), case_sensitive=False),
    help="Output format",
    callback=validate_output_format,
)
@click.pass_context
def cli(ctx, config_file, filename, output_format: OutputType | None):
    settings = load_settings(config_file)
    settings.data_filename = filename or settings.data_filename
    settings.output_format = output_format or settings.output_format
    ctx.obj = ContextObject(
        TimeTrackingService(settings),
        create_output_formatter(settings.output_format, force=(bool(output_format))),
    )


@cli.command("start")
@click.argument("project", required=True, type=str)
@click.argument("comment", required=False, type=str, default="")
@click.option("--tag", required=False, type=str, multiple=True)
@click.option(
    "--start_time",
    "-s",
    required=False,
    type=str,
    help="Start time expressed as HH:MM or HH:MM:SS in 24-hour format or AM/PM",
    default=None,
    callback=validate_time,
)
@pass_context_object
def start(context: ContextObject, project, comment, tag, start_time):
    """start tracking work on a project"""
    time_entry = context.tts.start_tracking(project, comment=comment, tags=set(tag), start_time=start_time)
    context.output.single_entry_output(time_entry)


@cli.command("stop")
@click.argument("comment", required=False, type=str, default="")
@click.option(
    "--stop_time",
    "-s",
    callback=validate_time,
    required=False,
    type=str,
    default=None,
    help="Stop/End time expressed as HH:MM or HH:MM:SS in 24-hour format or AM/PM",
)
@pass_context_object
def stop(context: ContextObject, comment, stop_time):
    """stop tracking work on a project"""
    if comment:
        time_entry = context.tts.get_active_entry()
        time_entry.comment = comment
        time_entry = context.tts.update_entry(time_entry)
    time_entry = context.tts.stop_tracking(stop_time=stop_time)
    context.output.single_entry_output(time_entry)


@cli.command("status")
@pass_context_object
def status(context: ContextObject):
    """displays currently tracked (if active)"""
    time_entry = context.tts.get_active_entry()
    context.output.single_entry_output(time_entry)


@cli.command("edit")
@click.argument("id", required=True, type=str)
@pass_context_object
def edit_entry(context: ContextObject, id):
    """edit a time entry using the system editor"""
    try:
        entry = context.tts.get_entry_by_partial_id(id)
    except KeyError as e:
        raise click.ClickException(f"No entry found with id {id}") from e
    except IndexError as e:
        raise click.ClickException(f"Multiple records found starting with id {id}") from e
    try:
        context.output.single_entry_output(context.tts.edit_entry(entry.id))
    except EditorServiceError as e:
        raise click.ClickException(f"Error editing entry: {str(e)}") from e


@cli.command("delete", aliases=["del", "rm"])
@click.argument("id", required=True, type=str)
@pass_context_object
def delete_entry(context: ContextObject, id):
    """delete a time entry"""
    try:
        entry = context.tts.get_entry_by_partial_id(id)  # Get the entry to delete
    except KeyError as e:
        raise click.ClickException(f"No entry found with id {id}") from e
    except IndexError as e:
        raise click.ClickException(f"Multiple records found starting with id {id}") from e
    entry = context.tts.delete_entry(entry.id)
    context.output.single_entry_output(entry)


@cli.command("list", aliases=["ls"])
@click.argument(
    "time_period",
    required=False,
    type=click.Choice(["today", "yesterday", "week", "month", "all", ""]),
    default="",
)
@click.option(
    "--start_date",
    type=Datetime(format="%Y-%m-%d"),
    help="Start date in format YYYY-MM-DD",
)
@click.option("--end_date", type=Datetime(format="%Y-%m-%d"), help="End date in format YYYY-MM-DD")
@click.option("--tag", multiple=True)
@click.option("--project", multiple=True)
@click.option("--format")
@pass_context_object
def list_entries(context: ContextObject, time_period, start_date, end_date, tag, project, format):
    """display list of time entries for a time period"""
    filter = EntryListFilter(
        time_period=time_period,
        start_date=start_date,
        end_date=end_date,
        tags=set(tag),
        projects=set(project),
        output_format=format,
    )
    time_list = context.tts.list_entries(filter=filter)
    context.output.multiple_entries_output(time_list)



from typing import Any

import click

from .output import OutputType


def validate_output_format(ctx: click.Context, param: Any, value: str | None) -> OutputType | None:
    """Validation callback for the output format option."""
    if value is not None:
        try:
            return OutputType(value.lower())
        except ValueError as e:
            raise click.BadParameter(f"Invalid output format: {value}") from e
    return None



import click
import pytest

from ..output_utils import validate_output_format


def test_validate_output_format():
    assert validate_output_format(None, None, "json") == "json"
    with pytest.raises(click.BadParameter):
        validate_output_format(None, None, "invalid")
    assert validate_output_format(None, None, None) is None
    assert validate_output_format(None, None, "YAML") == "yaml"
    with pytest.raises(click.BadParameter):
        validate_output_format(None, None, "INVALID")
